/*
 * Jason Spashett 2003.
 *
 * A simple Aho-Corasick virus detector.
 */

#include <windows.h>



#include <stdexcept>
#include <string>
#include <iostream>
#include <fstream>
#include <list>
#include <queue>
#include <cassert>
#include <ctime>
#include <cmath>
#include <iomanip>

using namespace std;


#include "scanner.h"


struct Session
{
	Scanner		ScanObj;
	long int	Detections;
	long int	InfectedFiles;
	long int	FilesFailedToOpen;
	long int	FilesScanned;
	bool		ShowFilesScanned;
	Session( )
	{
		Detections = 0;
		InfectedFiles = 0;
		FilesFailedToOpen = 0;
		FilesScanned = 0;
		ShowFilesScanned = 0;
	}
	void LoadDatabase( istream &database_stream )
	{
		ScanObj.LoadDatabase( database_stream );
	}
};


void show_activity()
{
	const char * const pic = "|/-\\";
	static int index = 0;

	// Straight to the console
	HANDLE out_handle = GetStdHandle( STD_ERROR_HANDLE );
	DWORD	n_written;
	WriteConsole(out_handle, pic + index, 1, &n_written, NULL);
	WriteConsole(out_handle, "\r", 1, &n_written, NULL);
	index = (index + 1 ) % sizeof pic;
}


void check_file( Session &session, string file_name )
{
	BlockBuffer	b;
	unsigned char buffer[16384];
	b.data = buffer;

	if ( session.ShowFilesScanned == true )
		cout << "Checking " << file_name << endl;

	fstream i(file_name.c_str(), ios_base::in | ios_base::binary );
	if ( i.is_open() )
	{
		session.FilesScanned++;
		session.ScanObj.Reset();
		bool found = false;
		while ( i.fail() == false )
		{
			b.i = 0;
			i.read( (char*)buffer, sizeof buffer );
			b.n = i.gcount();

			bool ret;
			list<const Pattern*> *p;

			while ( (ret = session.ScanObj.Find( &b, &p )) == true)
			{
				list<const Pattern*>::iterator i;

				for (i=p->begin(); i!= p->end(); ++i)
				{
					cout << "Detected: " << (*i)->m_Name << endl
					<< "\t in " << file_name << endl;
					session.Detections++;
					found = true; // Say we have found at least one.
				}
			}
		}
		if ( found == true ) // At least one was detected
		{
			session.InfectedFiles++;
		}
	}
	else
	{
		cerr << "Can't open: " << file_name << endl;
		session.FilesFailedToOpen++;
	}
}


void find_files(Session &session, const string &directory, const string &search_spec)
{
	WIN32_FIND_DATA find_data;
	HANDLE find_handle;

	clog << "find_files(" << directory << ", " << search_spec << ");" << endl;

	// Search for plain files first
	find_handle = FindFirstFile( (directory + search_spec).c_str(), &find_data );
	BOOL found = find_handle != INVALID_HANDLE_VALUE;
	while ( found )
	{
		show_activity(); // Show activity here, incase the search spec doesn't match many files at all.

		if ( strcmp( find_data.cFileName, "." ) && strcmp( find_data.cFileName, ".." ) )
		{
			if ( (find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0 ) // A file?
			{
				string file_name(directory );
				file_name += "/";
				file_name.append( find_data.cFileName );
				check_file( session, file_name );
			}
		}
		found = FindNextFile(find_handle, &find_data);
	}
	FindClose(find_handle);

	// Search for directories to traverse now
	find_handle = FindFirstFile( (directory + "/*").c_str(), &find_data );
	found = find_handle != INVALID_HANDLE_VALUE;
	while ( found )
	{
		if ( strcmp( find_data.cFileName, "." ) && strcmp( find_data.cFileName, ".." ) )
		{
			if ( (find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0 ) // A file?
			{
				string file_name(directory );
				file_name += "/";
				file_name.append( find_data.cFileName );
				find_files(session, file_name, search_spec);
			}
		}
		found = FindNextFile(find_handle, &find_data);
	}
	FindClose(find_handle);

}

int main(int argc, char *argv[])
{
	time_t	start, finish;
	time(&start);

	Session	session;
	const char *signature_file = "viruses.db";

	// Switch log file off.
	clog.setstate( ios::badbit );

	if ( argc > 1 )
	{
		try
		{
			// Check for switches
			for (int i=1; i < argc; ++i)
			{
				if (argv[i][0] == '-' || argv[i][0] == '/') // A switch
				{
					switch ( tolower( (char) argv[i][1]) )
					{
						case 'd':
							clog.clear();	// Turn clog on.
						break;

						case 's':
							signature_file = &argv[i][2];
							break;

						case 'l':
							session.ShowFilesScanned = true;
							break;

						default:
							throw invalid_argument( "Unknown option" );
							break;
					}
				}
			}

			// Load signature database
			ifstream signature_stream( signature_file );
			if ( !signature_stream.is_open() )
				throw runtime_error("Can't open signature file: " + string(signature_file) );
			session.LoadDatabase( signature_stream );
			signature_stream.close();

			// check for file arguments
			for (int i=1; i < argc; ++i)
			{
				string arg;
				if (argv[i][0] != '-' && argv[i][0] != '/')
				{
					if ( argv[i][strlen(argv[i])-1] == ':' )
					{
						arg = string(argv[i]) + "/"; // Add a slash to scan from the root then.
					}
					else
					{
						// Get current directory
						char path_buf[MAX_PATH], *name;
						GetFullPathName( argv[i], sizeof path_buf, path_buf, &name );
						arg = path_buf;
					}

					// Is it a directory?
					DWORD file_attributes = GetFileAttributes(arg.c_str());
					if ( INVALID_FILE_ATTRIBUTES != file_attributes &&
						(FILE_ATTRIBUTE_DIRECTORY & file_attributes ) != 0 )
					{
						char last_char = arg.at( arg.length()-1 );
						if ( last_char != '\\' && last_char != '/')
							arg += "/";
						arg += '*';
					}

					string path;
					string search_spec;

					// A search spec?
					if ( arg.find_last_of("*?") == arg.npos )
					{
						// Then it's a single file
						check_file(session, arg);
					}
					else
					{
						// Find the search spec part
						string::size_type pos = arg.find_last_of("\\/");
						if (pos == arg.npos)
							pos = 0;

						path = arg.substr(0, pos); // Extract the path part
						search_spec = arg.substr(pos); // Get the search spec part
						find_files(session, path, search_spec);
					}
				}
			}
			cout << endl;
			cout << "Viruses detected: " << session.Detections << endl;
			cout << "Files infected: " << session.InfectedFiles << endl;
			cout << "Files scanned: " << session.FilesScanned << endl;
			cout << "Files failed to open: " << session.FilesFailedToOpen << endl;
			cout << endl;
			cout << "Duplicate signatures in virus database: " << session.ScanObj.GetDuplicates() << endl;
		}
		catch (exception &e)
		{
			cerr << "Exception: " << e.what() << endl;
			throw;
		}
	}
	else
	{
		cout << argv[0] << " [-l] [-s<signature-file>] [-d] <file-spec>" << endl
		<< endl
		<< "\t" << "-l                  show files scanned" << endl
		<< "\t" << "-d                  debug output" << endl
		<< "\t" << "-s<signature-file>  path to signature file." << endl;
	}

	time(&finish);
	double elapsed = difftime( finish, start );

	double seconds = floor( fmod(elapsed, 60) );
	double minutes = floor( fmod(elapsed / 60, 60) );
	double hours = floor( elapsed / 3600 );

	cout << setfill('0');
	cout
		<< "Elapsed time: "
		<< hours
		<< ":"
		<< setw(2) << minutes
		<< "."
		<< setw(2) << seconds
		<< " (" << elapsed << " seconds" << ")"
		<< endl;

	return EXIT_SUCCESS;
}
